@@ -5,7 +5,6 @@
  * published by the Free Software Foundation.
  */
 #include <assert.h>				/* assert */
-#include <ctype.h>				/* tolower */
 #include <string.h>				/* memcmp, str* */
 
 #include <libipset/linux_ip_set.h>		/* IPSET_CMD_* */
@@ -25,91 +24,91 @@ const struct ipset_commands ipset_commands[] = {
 
 	{	/* c[reate], --create, n[ew], -N */
 		.cmd = IPSET_CMD_CREATE,
-		.name = { "create", "new" },
+		.name = { "create", "new", "-N" },
 		.has_arg = IPSET_MANDATORY_ARG2,
 		.help = "SETNAME TYPENAME [type-specific-options]\n"
 			"        Create a new set",
 	},
 	{	/* a[dd], --add, -A  */
 		.cmd = IPSET_CMD_ADD,
-		.name = { "add", NULL },
+		.name = { "add", "-A", NULL },
 		.has_arg = IPSET_MANDATORY_ARG2,
 		.help = "SETNAME ENTRY\n"
 			"        Add entry to the named set",
 	},
 	{	/* d[el], --del, -D */
 		.cmd = IPSET_CMD_DEL,
-		.name = { "del", NULL },
+		.name = { "del", "-D", NULL },
 		.has_arg = IPSET_MANDATORY_ARG2,
 		.help = "SETNAME ENTRY\n"
 			"        Delete entry from the named set",
 	},
 	{	/* t[est], --test, -T */
 		.cmd = IPSET_CMD_TEST,
-		.name = { "test", NULL },
+		.name = { "test", "-T", NULL },
 		.has_arg = IPSET_MANDATORY_ARG2,
 		.help = "SETNAME ENTRY\n"
 			"        Test entry in the named set",
 	},
 	{	/* des[troy], --destroy, x, -X */
 		.cmd = IPSET_CMD_DESTROY,
-		.name = { "destroy", "x" },
+		.name = { "destroy", "x", "-X" },
 		.has_arg = IPSET_OPTIONAL_ARG,
 		.help = "[SETNAME]\n"
 			"        Destroy a named set or all sets",
 	},
 	{	/* l[ist], --list, -L */
 		.cmd = IPSET_CMD_LIST,
-		.name = { "list", NULL },
+		.name = { "list", "-L", NULL },
 		.has_arg = IPSET_OPTIONAL_ARG,
 		.help = "[SETNAME]\n"
 			"        List the entries of a named set or all sets",
 	},
 	{	/* s[save], --save, -S */
 		.cmd = IPSET_CMD_SAVE,
-		.name = { "save", NULL },
+		.name = { "save", "-S", NULL },
 		.has_arg = IPSET_OPTIONAL_ARG,
 		.help = "[SETNAME]\n"
 			"        Save the named set or all sets to stdout",
 	},
 	{	/* r[estore], --restore, -R */
 		.cmd = IPSET_CMD_RESTORE,
-		.name = { "restore", NULL },
+		.name = { "restore", "-R", NULL },
 		.has_arg = IPSET_NO_ARG,
 		.help = "\n"
 			"        Restore a saved state",
 	},
 	{	/* f[lush], --flush, -F */
 		.cmd = IPSET_CMD_FLUSH,
-		.name = { "flush", NULL },
+		.name = { "flush", "-F", NULL },
 		.has_arg = IPSET_OPTIONAL_ARG,
 		.help = "[SETNAME]\n"
 			"        Flush a named set or all sets",
 	},
 	{	/* ren[ame], --rename, e, -E */
 		.cmd = IPSET_CMD_RENAME,
-		.name = { "rename", "e" },
+		.name = { "rename", "e", "-E" },
 		.has_arg = IPSET_MANDATORY_ARG2,
 		.help = "FROM-SETNAME TO-SETNAME\n"
 			"        Rename two sets",
 	},
 	{	/* sw[ap], --swap, w, -W */
 		.cmd = IPSET_CMD_SWAP,
-		.name = { "swap", "w" },
+		.name = { "swap", "w", "-W" },
 		.has_arg = IPSET_MANDATORY_ARG2,
 		.help = "FROM-SETNAME TO-SETNAME\n"
 			"        Swap the contect of two existing sets",
 	},
 	{	/* h[elp, --help, -H */
 		.cmd = IPSET_CMD_HELP,
-		.name = { "help", NULL },
+		.name = { "help", "-h", "-H" },
 		.has_arg = IPSET_OPTIONAL_ARG,
 		.help = "[TYPENAME]\n"
 			"        Print help, and settype specific help",
 	},
 	{	/* v[ersion], --version, -V */
 		.cmd = IPSET_CMD_VERSION,
-		.name = { "version", NULL },
+		.name = { "version", "-v", "-V" },
 		.has_arg = IPSET_NO_ARG,
 		.help = "\n"
 			"        Print version information",
@@ -128,32 +127,55 @@ const struct ipset_commands ipset_commands[] = {
 bool
 ipset_match_cmd(const char *arg, const char * const name[])
 {
-	size_t len;
+	size_t len, skip = 0;
+	int i;
 
 	assert(arg);
 	assert(name && name[0]);
 
-	/* Ignore (two) leading dashes */
-	if (arg[0] == '-')
-		arg++;
-	if (arg[0] == '-')
-		arg++;
+	/* Ignore two leading dashes */
+	if (arg[0] == '-' && arg[1] == '-')
+		skip = 2;
 
 	len = strlen(arg);
+	if (len <= skip || (len == 1 && arg[0] == '-'))
+		return false;
 
-	if (len > strlen(name[0]) || !len)
-		return false;
-	else if (len > 1 &&
-		 ((strncmp(arg, name[0], len) == 0) ||
-		  (name[1] != NULL && (strncmp(arg, name[1], len) == 0))))
-		return true;
-	else if (len != 1)
-		return false;
-	else
-		return tolower(arg[0]) == name[0][0] ||
-		       (name[1] != NULL && tolower(arg[0]) == name[1][0]);
+	for (i = 0; i < IPSET_CMD_ALIASES && name[i] != NULL; i++) {
+		/* New command name options */
+		if (STRNEQ(arg + skip, name[i], len - skip))
+			return true;
+	}
+	return false;
 }
 
+/* Used up so far
+ *
+ *	-A		add
+ *	-D		del
+ *	-E		rename
+ *	-f		-file
+ *	-F		flush
+ *	-h		help
+ *	-H		help
+ *	-L		list
+ *	-n		-name
+ *	-N		create
+ *	-o		-output
+ *	-r		-resolve
+ *	-R		restore
+ *	-s		-sorted
+ *	-S		save
+ *	-t		-terse
+ *	-T		test
+ *	-q		-quiet
+ *	-X		destroy
+ *	-v		version
+ *	-V		version
+ *	-W		swap
+ *	-!		-exist
+ */
+
 const struct ipset_envopts ipset_envopts[] = {
 	{ .name = { "-o", "-output" },
 	  .has_arg = IPSET_MANDATORY_ARG,	.flag = IPSET_OPT_MAX,
@@ -202,6 +224,14 @@ const struct ipset_envopts ipset_envopts[] = {
 		  "        When listing, list setnames and set headers\n"
 		  "        from kernel only.",
 	},
+	{ .name = { "-f", "-file" },
+	  .parse = ipset_parse_file,
+	  .has_arg = IPSET_MANDATORY_ARG,	.flag = IPSET_OPT_MAX,
+	  .help = "\n"
+		  "        Read from the given file instead of standard\n"
+		  "        input (restore) or write to given file instead\n"
+		  "        of standard output (list/save).",
+	},
 	{ },
 };
 
